#include <iostream>
#include <fstream>
#include <stdexcept>
#include <vector>
#include <stdint.h>
#include <climits>
#include <cassert>
#include <algorithm>
#include <sys/time.h>

#include "CL/opencl.h"

#include "setupGPU.h"
#include "memCNTRL.h"
#include "utilsPaPaRa.h"

int References =256;
//int KernelIterations = 1;
int Queries = 2180;


static double gettime(void )
{
    struct timeval ttime;
    gettimeofday(&ttime , 0);
    return ttime.tv_sec + ttime.tv_usec * 0.000001;
}


class timer {
    double m_st;
    
public:
    timer() : m_st(gettime()) {};
    
    double elapsed() const {
        return gettime() - m_st;
    }
    
    
};



typedef int score_t;
typedef unsigned short score_t_exp;
typedef unsigned int u32_de;


struct align_arrays {
    std::vector<score_t> s;
   // std::vector<score_t> si;
};


static score_t align_pvec_score( std::vector<int> &a, std::vector<unsigned int> &a_aux, const std::vector<uint8_t> &b, score_t mismatch_score, score_t match_cgap, score_t gap_open, score_t gap_extend, align_arrays &arr ) {

    if( arr.s.size() < a.size()  ) {
        arr.s.resize( a.size() );
    }
    
    const score_t aux_cgap = 0x1;
    
    const size_t band_width = a.size() - b.size();
    std::fill( arr.s.begin(), arr.s.end(), 0 );
    
    const score_t LARGE = 32000;
//     std::fill( arr.si.begin(), arr.si.end(), LARGE );
    arr.s.at(band_width+1) = LARGE;

    score_t min_score = LARGE;
    
    for( size_t ib = 0; ib < b.size(); ib++ ) {
        int bc = b[ib];
        
        score_t last_sl = LARGE;
        score_t last_sc = LARGE;
        score_t last_sdiag = 0.0;
        
        size_t astart = ib;
                
        score_t * __restrict s_iter = &arr.s[0];

        
        bool lastrow = ib == (b.size() - 1);
        
        last_sdiag = *s_iter;
        
        
        for( size_t ia = astart; ia <= ib + band_width; ++ia, ++s_iter ) {  
            score_t ac = a[ia];
            const bool cgap = a_aux[ia] == aux_cgap;
            
            // determine match or mis-match according to parsimony bits coming from the tree.
            bool is_match = ( ac & bc ) != 0;

            score_t sm = last_sdiag;
            
            if( !is_match ) {
                sm += mismatch_score;
            }
            
            last_sdiag = *(s_iter+1);

            score_t last_sc_OPEN; 
            score_t sl_score_stay;
            
            if( cgap ) {
                last_sc_OPEN = last_sc; 
                sl_score_stay = last_sl;
                sm += match_cgap;
            } else {
                last_sc_OPEN = last_sc + gap_open + gap_extend;  
                sl_score_stay = last_sl + gap_open;
            }
            
            score_t sl = std::min( sl_score_stay, last_sc_OPEN );
                        
            last_sl = sl;
            

            score_t su = last_sdiag + mismatch_score;
            
//             std::cout << "x: " << sm << " " << su << " " << sl << "  ";
            score_t sc = std::min( sm, std::min( su, sl ) );
            
            last_sc = sc;
            *s_iter = sc;
            
        
            if( lastrow ) {
                if( sc < min_score ) {
                    
                    min_score = sc;
                }
            }
           
        }
//         std::cout << "\n";
//         getchar();
    }

    return min_score;
    
}



class testbench {
    std::vector<std::vector <int> > m_ref_pvecs;
    std::vector<std::vector <unsigned int> > m_ref_aux;
    
    std::vector<std::vector <uint8_t> > m_qs_pvecs;

    
    void load_qs( std::istream &is ) {
        size_t num_qs;
        is >> num_qs;
        
        m_qs_pvecs.resize( num_qs );
        
        for( int i = 0; i < m_qs_pvecs.size(); ++i ) {
            size_t len;
            is >> len;
            while( isspace( is.get() ) && !is.eof() ) {}
            is.unget();
            
            if( is.eof() ) {
                throw std::runtime_error( "unexpected end of file" );
            }
            
            m_qs_pvecs[i].resize(len);
            is.read( (char*)&m_qs_pvecs[i][0], len );
//             std::cout << "qs: " << i << " " << m_qs_pvecs[i].size() << "\n";
        }
    }
    
    void load_ref( std::ifstream &is ) {
        size_t num_ref;
        is >> num_ref;
        m_ref_pvecs.resize( num_ref );
        m_ref_aux.resize( num_ref );
        
        for( int i = 0; i < m_ref_pvecs.size(); ++i ) {
            size_t len;
            is >> len;
            while( isspace( is.get() ) && !is.eof() ) {}
            is.unget();
            
            if( is.eof() ) {
                throw std::runtime_error( "unexpected end of file" );
            }
            
            m_ref_pvecs[i].resize(len);
            is.read( (char*)&m_ref_pvecs[i][0], len * sizeof(int) );
            
            m_ref_aux[i].resize(len);
            is.read( (char*)&m_ref_aux[i][0], len * sizeof(unsigned int) );
            
//             std::cout << "ref: " << i << " " << m_ref_pvecs[i].size() << "\n";
        }
        
    }
    template<typename T, size_t min_len>
    static void trim_ref( std::vector<T> &v ) {
        
    	const size_t max_len = min_len;
        while( v.size() < min_len ) {
            std::vector<T> v_new = v;
            
            
            std::copy( v.begin(), v.end(), std::back_inserter(v_new) );
            
            v.swap(v_new);
            
        }
        

        if( v.size() > max_len ) {
            v.resize( max_len );
	    //v.erase( v.begin(), v.begin() + v.size() - max_len );
        }
        
        
    }   
 
    template<typename cont>
    void duplicate( cont &c ) {
	assert( !c.empty() );
    	cont tmp = c;
	c.insert( c.end(), tmp.begin(), tmp.end() );
    }

public:  
    testbench( const char *qs_name, const char *ref_name ) {
        {
            std::ifstream is( qs_name );
            load_qs( is );
        }
        {
            std::ifstream is( ref_name );
            load_ref( is );
        }

	while( m_ref_pvecs.size() < References ) {
		duplicate( m_ref_pvecs );
		duplicate( m_ref_aux );

	}

	std::srand(1234);
	std::random_shuffle( m_ref_pvecs.begin(), m_ref_pvecs.end() );
	std::srand(1234);
	std::random_shuffle( m_ref_aux.begin(), m_ref_aux.end() );
        
        
        // reduce size of testset
        size_t num_ref = References; // JUMP
        size_t num_qs = Queries;
        
        num_ref = std::min( num_ref, m_ref_pvecs.size() );
        num_qs = std::min( num_qs, m_qs_pvecs.size() );
        
        m_ref_pvecs.resize(num_ref);
        m_ref_aux.resize(num_ref);
        m_qs_pvecs.resize(num_qs);
        
	//const size_t len = 1200;
	//std::for_each( m_ref_pvecs.begin(), m_ref_pvecs.end(), &trim_ref<int,len> );
        //std::for_each( m_ref_aux.begin(), m_ref_aux.end(), &trim_ref<unsigned int,len> );
	
        int ct = 0;
	for( size_t i = 0; i < m_ref_aux.size(); ++i ) {
 	if( m_ref_aux[i].back() != 0 ) { ct++; }
	}
	std::cout << "\n\nCGAP REFS:\t\t"<<ct << "\n";


    }
    
    
    void run_seq() {
        const int score_gap_open = 1;
        const int score_gap_extend = 1;
        const int score_mismatch = 1;
        const int score_match_cgap = 4;
        
        
        
        uint64_t ncup = 0;
        
        
        /******************************************/
	/* Acceleration attempt implemented by me */
        /******************************************/

        /* Setting up GPU.. */

	int i;
	
	cl_int ciErrNum;

   	cl_platform_id* myPlatform = select_NVIDIA_platform();

    	cl_device_id * myDevice = select_GPU_device(myPlatform);

    	cl_context myContext= create_context(myDevice);

    	cl_command_queue myCommandQueue = create_command_queue (myDevice, myContext);

    	int kernel_size;



	/* Generic Kernel *
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Generic-Kernel.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLGenericKernel");
	/* */

	/* Compressed Kernel 12*/
	char * kernel_code = load_kernel_code ("PaPaRa-OpenCL-Unfold12-Kernel-Compressed.cl", &kernel_size);
	cl_program myProgram = create_and_build_program(myDevice, myContext, kernel_code, kernel_size);
	cl_kernel myKernel = create_kernel(myProgram, "PaPaRaOpenCLUnfold12KernelCompressed");
	/* */

	


/******************************************************************************************************************/
/* OPTIMIZED IMPLEMENTATION */
/* KERNEL NAME: GENERIC*/

/* *
const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";


// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";


// Shared Memory 
const size_t shared_memory_size = 4;//KBs
const size_t templineSMaxLen = (shared_memory_size * 1024)/4; // integers




// Blocked Version Parameters
const size_t BLOCKlengthCOLUMNS = templineSMaxLen / RefPadBloLen;
std::cout<<"Block size in columns:\t\t"<<BLOCKlengthCOLUMNS<<"\n";
const size_t tempcolGlen  = QueNum * QueLenMax * RefPadBloLen;
const size_t templineGlen = QueNum * (QueLenMax + BLOCKlengthCOLUMNS) * RefPadBloLen;
const size_t templineSlen = QueNum * BLOCKlengthCOLUMNS * RefPadBloLen;




int total = RefLen;
int partF = QueLenMax; 
int partL = QueLenMax;
int partM = total - partF - partL;
int blockiterations = partM / BLOCKlengthCOLUMNS;
int sitesinblocks = blockiterations * BLOCKlengthCOLUMNS;
int partLext = partM - sitesinblocks;
    partL = partL + partLext;
int lastblock_refstart = total - partL;






// Arguments
const size_t ArgVecLen = QueNum + 8; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = QueLenMax;
ArgVec[5] = BLOCKlengthCOLUMNS;
ArgVec[6] = blockiterations;
ArgVec[7] = lastblock_refstart;

for(size_t i = 8; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-8].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);

//cl_mem SLVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
//cl_mem SCVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
cl_mem SLSCVecDev = allocateDeviceMemory_u32 (tempcolGlen, myContext);
cl_mem BLOCKAW_lineDev = allocateDeviceMemory_32 (templineGlen , myContext);
//cl_mem BLOCK_lineDev = allocateDeviceMemory_32 (templineSlen, myContext);


// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &ResVecDev);

//clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &SLVecDev);
//clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &SCVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &BLOCKAW_lineDev);
clSetKernelArg(myKernel, 5, sizeof(int)*templineSMaxLen, 0);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &SLSCVecDev);
clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;
timer NikosTimerN;
clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);
std::cout << "\nGPU Time:\t\t" << NikosTimerN.elapsed() <<"\n\n";


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);


/* */

/******************************************************************************************************************/



/******************************************************************************************************************/
/* COMPRESSED IMPLEMENTATION */
/* KERNEL NAME: COMPRESSED*/

/* */


const size_t RefNum = References; 
const size_t QueNum = Queries;
const size_t RefLen = m_ref_pvecs[0].size();
int temp_quelen_max = 0;for(size_t i=0;i<m_qs_pvecs.size();i++) {if(temp_quelen_max<m_qs_pvecs[i].size()){temp_quelen_max=m_qs_pvecs[i].size();}}
const size_t QueLenMax = temp_quelen_max;
std::cout << "Number of References:\t\t"<< RefNum << "\nNumber of Queries:\t\t" << QueNum << "\nReference Length:\t\t" << RefLen << "\nMaximum Query Length:\t\t" << QueLenMax <<"\n";

float globalsizeINTS=0;

// Query Organization: Placement one after the other in blocks of size QueLenMax and zero filling.
const size_t QueVecLen = QueNum * QueLenMax; // number of unsigned integers in the QueVec
unsigned int * QueVec = (unsigned int *) malloc(sizeof(unsigned int)*QueVecLen);
for(size_t i=0;i<QueNum;i++){putSequenceQue(i,m_qs_pvecs[i],QueLenMax,QueVec);}



// Inter-Reference Memory Organization
const size_t RefPadBloLen = get_next32mult (RefNum);
std::cout << "Reference Block Size:\t\t"<<RefPadBloLen<<"\n\n";
const size_t InterRefVecLen = RefPadBloLen * RefLen;
const float InterRefVecGBs = ((float)InterRefVecLen*32.0)/8.0/1024.0/1024.0/1024.0;
//int * InterRefVec_a = (int *) malloc(sizeof(int)*InterRefVecLen);
//unsigned int * InterRefVec_a_aux = (unsigned int *) malloc(sizeof(unsigned int)*InterRefVecLen);
//for(size_t i=0;i<RefNum;i++){putSequenceRef_InterInt(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,InterRefVec_a,RefPadBloLen);}
std::cout << "Reference Block GBs:\t\t"<<InterRefVecGBs*2<<"GBs\n\n";


// Shared Memory 
const size_t shared_memory_size = 12;//KBs
const size_t templineSMaxLen = (shared_memory_size * 1024)/4; // integers




// Blocked Version Parameters
const size_t BLOCKlengthCOLUMNS = templineSMaxLen / RefPadBloLen;
std::cout<<"Block size in columns:\t\t"<<BLOCKlengthCOLUMNS<<"\n";
const size_t tempcolGlen  = QueNum * QueLenMax * RefPadBloLen;
const size_t templineGlen = QueNum * (QueLenMax + BLOCKlengthCOLUMNS) * RefPadBloLen;
const size_t templineSlen = QueNum * BLOCKlengthCOLUMNS * RefPadBloLen;




int total = RefLen;
int partF = QueLenMax; 
int partL = QueLenMax;
int partM = total - partF - partL;
int blockiterations = partM / BLOCKlengthCOLUMNS;
std::cout<<"Iterations:\t\t"<<blockiterations<<"\n";
int sitesinblocks = blockiterations * BLOCKlengthCOLUMNS;
int partLext = partM - sitesinblocks;
    partL = partL + partLext;
int lastblock_refstart = total - partL;

std::cout<<"Limit1:"<< partF <<"-- Limit2:"<<lastblock_refstart<<"LastL:"<<partL<<"Check:"<<partF+blockiterations*BLOCKlengthCOLUMNS+partL;

// Inter-Reference Memory Middle Part Compression: 4 ref entries per int
const size_t compression_effort = 6;
const size_t compressed_vector_length = ((lastblock_refstart - partF)/compression_effort)*RefPadBloLen;
std::cout<<"Compressed Vector Length:"<<compressed_vector_length;
int * CompRefMidVec = (int *) malloc(sizeof(int)*compressed_vector_length);
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterIntComp(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,partF, lastblock_refstart,CompRefMidVec,RefPadBloLen);}


// Merge uncompressed and compressed values

const size_t RefVecLen_Combined = (partF + partL + blockiterations*BLOCKlengthCOLUMNS/compression_effort)*RefPadBloLen;
std::cout<<"Combined length:"<<RefVecLen_Combined;
int * RefVecCombined = (int *) malloc (sizeof(int)*RefVecLen_Combined);
for(size_t i=0;i<RefNum;i++){putSequenceRef_CombinedS1(i,m_ref_pvecs[i], m_ref_aux[i],partF,RefVecCombined,RefPadBloLen);}
std::cout<<"1:"<< partF <<"\n2:"<<blockiterations<<"\n3:"<<sitesinblocks<<"\n4:"<<lastblock_refstart;
for(size_t i=0;i<RefNum;i++){putSequenceRef_CombinedS2(i,m_ref_pvecs[i], m_ref_aux[i], partL, RefVecCombined, RefPadBloLen, partF, blockiterations, BLOCKlengthCOLUMNS/compression_effort,lastblock_refstart);}
for(size_t i=0;i<RefNum;i++){putSequenceRef_InterIntCompCombined(i,m_ref_pvecs[i], m_ref_aux[i],RefLen,partF, lastblock_refstart,RefVecCombined,RefPadBloLen,partF);}

//unsigned int quelenmax, unsigned int blockiterations, unsigned int blocksizecompressed, unsigned int startpos);


// Arguments
const size_t ArgVecLen = QueNum + 9; // Set to RefLen for kernel testing
int * ArgVec = (int *) malloc(sizeof(int)*ArgVecLen);
ArgVec[0] = QueNum;
ArgVec[1] = RefNum;
ArgVec[2] = RefLen;
ArgVec[3] = RefPadBloLen;
ArgVec[4] = QueLenMax;
ArgVec[5] = BLOCKlengthCOLUMNS;
ArgVec[6] = blockiterations;
ArgVec[7] = lastblock_refstart;
ArgVec[8] = lastblock_refstart - (QueLenMax + blockiterations * BLOCKlengthCOLUMNS / compression_effort);

std::cout<< "previous:"<<ArgVec[7]<<"-current:"<<ArgVec[8]<<"\n";

for(size_t i = 9; i< ArgVecLen; i++)
{
	ArgVec[i] = m_qs_pvecs[i-9].size();
}
const size_t InterIntVecLen = QueNum *  InterRefVecLen;
const size_t ResVecLen = QueNum * RefPadBloLen;

// GPU Memory Allocation
cl_mem QueVecPin = allocatePinnedHostMemory_u32 (QueVecLen, myContext, myCommandQueue);
QueVecPin = initializePinnedHostMemory_u32_Sequence (QueVecLen, QueVecPin, myCommandQueue, QueVec);
cl_mem QueVecDev = allocateDeviceMemory_u32 (QueVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_u32 (QueVecLen, QueVecPin, QueVecDev, myCommandQueue);

globalsizeINTS+=QueVecLen;

//cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (InterRefVecLen, myContext, myCommandQueue);
//InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (InterRefVecLen, InterRefVecPin_a, myCommandQueue, InterRefVec_a);
//cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (InterRefVecLen, myContext);
//copyPinnedHostMemoryToDeviceMemory_32 (InterRefVecLen, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

cl_mem InterRefVecPin_a = allocatePinnedHostMemory_32 (RefVecLen_Combined, myContext, myCommandQueue);
InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (RefVecLen_Combined, InterRefVecPin_a, myCommandQueue, RefVecCombined);
cl_mem InterRefVecDev_a = allocateDeviceMemory_32 (RefVecLen_Combined, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (RefVecLen_Combined, InterRefVecPin_a, InterRefVecDev_a, myCommandQueue);

globalsizeINTS+=RefVecLen_Combined;

//cl_mem InterRefVecCompPin_a = allocatePinnedHostMemory_32 (compressed_vector_length, myContext, myCommandQueue);
//InterRefVecPin_a = initializePinnedHostMemory_32_Sequence (compressed_vector_length, InterRefVecCompPin_a, myCommandQueue, CompRefMidVec);
//cl_mem InterRefVecCompDev_a = allocateDeviceMemory_32 (compressed_vector_length, myContext);
//copyPinnedHostMemoryToDeviceMemory_32 (compressed_vector_length, InterRefVecCompPin_a, InterRefVecCompDev_a, myCommandQueue);

cl_mem ArgVecPin = allocatePinnedHostMemory_32 (ArgVecLen, myContext, myCommandQueue);
ArgVecPin = initializePinnedHostMemory_32_Sequence (ArgVecLen, ArgVecPin, myCommandQueue, ArgVec);
cl_mem ArgVecDev = allocateDeviceMemory_32 (ArgVecLen, myContext);
copyPinnedHostMemoryToDeviceMemory_32 (ArgVecLen, ArgVecPin, ArgVecDev, myCommandQueue);

globalsizeINTS+=ArgVecLen;

cl_mem ResVecPin = allocatePinnedHostMemory_32 (ResVecLen, myContext, myCommandQueue);
cl_mem ResVecDev = allocateDeviceMemory_32 (ResVecLen, myContext);

globalsizeINTS+=ResVecLen;

//cl_mem SLVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
//cl_mem SCVecDev = allocateDeviceMemory_32 (tempcolGlen , myContext);
cl_mem SLSCVecDev = allocateDeviceMemory_u32 (tempcolGlen, myContext);
globalsizeINTS+=tempcolGlen;
cl_mem BLOCKAW_lineDev = allocateDeviceMemory_32 (templineGlen , myContext);
globalsizeINTS+=templineGlen;
//cl_mem BLOCK_lineDev = allocateDeviceMemory_32 (templineSlen, myContext);

globalsizeINTS = globalsizeINTS*4.0/1024.0/1024.0/1024.0;
std::cout<<"\n\nGlobalsize:\t\t"<<globalsizeINTS<<" or "<<globalsizeINTS*1024.0<<"\n\n";

// Lunch GPU kernel
clSetKernelArg(myKernel, 0, sizeof(cl_mem), (void *) &QueVecDev);
clSetKernelArg(myKernel, 1, sizeof(cl_mem), (void *) &InterRefVecDev_a);
clSetKernelArg(myKernel, 2, sizeof(cl_mem), (void *) &ArgVecDev);
clSetKernelArg(myKernel, 3, sizeof(cl_mem), (void *) &ResVecDev);

//clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &SLVecDev);
//clSetKernelArg(myKernel, 5, sizeof(cl_mem), (void *) &SCVecDev);
clSetKernelArg(myKernel, 4, sizeof(cl_mem), (void *) &BLOCKAW_lineDev);
clSetKernelArg(myKernel, 5, sizeof(int)*templineSMaxLen, 0);
clSetKernelArg(myKernel, 6, sizeof(cl_mem), (void *) &SLSCVecDev);
//clSetKernelArg(myKernel, 7, sizeof(cl_mem), (void *) &InterRefVecCompDev_a);
clFinish(myCommandQueue);

size_t WorkSizeGlobal[] = {RefPadBloLen * QueNum};
size_t WorkSizeLocal[] = {RefPadBloLen};

cl_event event;
timer NikosTimerN;


clEnqueueNDRangeKernel(myCommandQueue, myKernel, 1, NULL, WorkSizeGlobal, WorkSizeLocal, 0, NULL, &event);
clFinish(myCommandQueue);
double NikosTimerNelapsed = NikosTimerN.elapsed();
std::cout << "\nGPU Time:\t\t" << NikosTimerNelapsed <<"\n\n";


// Get scores back
int * ResVec = copyDeviceMemoryToPinnedHostMemory_32 (ResVecLen, ResVecDev, ResVecPin, myCommandQueue);
clFinish(myCommandQueue);


/* */

/******************************************************************************************************************/






/* SIMON'S REFERENCE TEST */

unsigned int score_ref;
//for(size_t i=0;i<1;i++)
//{
//score_ref = align_pvec_score( m_ref_pvecs[1], m_ref_aux[1], m_qs_pvecs[198], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
//std::cout << "Simon score for 0 0 0     " << score_ref <<"\n";
//}
unsigned int * ResVecRef = (unsigned int *) malloc (sizeof(unsigned int)*ResVecLen);
size_t y = 0, yy=0;
unsigned int fillzeros = RefPadBloLen - RefNum;
unsigned int ncups = 0;
align_arrays arr;

timer SimonTimerN;
for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) 
{
 	for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) 
	{          
                unsigned int score = align_pvec_score( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		//ResVec[y] = align_pvec_scoreBLOCKED( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		ResVecRef[y++] = score;

		ncups += (m_ref_pvecs[i].size() - m_qs_pvecs[j].size())*m_qs_pvecs[j].size();	
        }  
	for(size_t yy=0;yy<fillzeros;yy++)
	{
		//ResVec[y] = 0;
		ResVecRef[y++] = 0;
 	}          
}
std::cout << "\nCPU Time:\t\t" << SimonTimerN.elapsed() <<"\n\n";
std::cout << "\n GPU GCUPS = \t\t"<< (ncups/NikosTimerNelapsed) * 1000000000;


unsigned int totalcounter=0;

//std::cout << "GNcupsCPU: " << ncups / (NikosTime * 1e9) << "\n";
//std::cout << "GNcupsCPU: " << ncups / (SimonTimerN.elapsed() * 1e9) << "\n";
for(size_t i=0;i<ResVecLen;i++)
{
	
	if(ResVec[i]!=ResVecRef[i])
	{
		std::cout<<i<<":"<<ResVec[i]<<" and " << ResVecRef[i]<<"\n";
		totalcounter++;
	}
}

std::cout<<"\n\nIt is:"<<totalcounter<<" out of:"<<ResVecLen<<"\n\n";


/*
y = 0, yy=0;
timer NIKOSTimerN;
for( size_t j = 0; j < m_qs_pvecs.size(); ++j ) 
{
 	for( size_t i = 0; i < m_ref_pvecs.size(); ++i ) 
	{          
                //unsigned int score = align_pvec_score( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		ResVec[y++] = align_pvec_scoreBLOCKED_OPTIMIZED_3( m_ref_pvecs[i], m_ref_aux[i], m_qs_pvecs[j], score_mismatch, score_match_cgap, score_gap_open, score_gap_extend, arr );
		//ResVecRef[y++] = score;	
        }  
	for(size_t yy=0;yy<fillzeros;yy++)
	{
		ResVec[y++] = 0;
		//ResVecRef[y++] = 0;
 	}          
}
std::cout << "\nCPU Time NEW:\t\t" << NIKOSTimerN.elapsed() <<"\n\n";


totalcounter=0;

//std::cout << "GNcupsCPU: " << ncups / (NikosTime * 1e9) << "\n";
//std::cout << "GNcupsCPU: " << ncups / (SimonTimerN.elapsed() * 1e9) << "\n";
for(size_t i=0;i<ResVecLen;i++)
{
	
	if(ResVec[i]!=ResVecRef[i])
	{
		std::cout<<i<<":"<<ResVec[i]<<" and " << ResVecRef[i]<<"\n";
		totalcounter++;
	}
}

std::cout<<"\n\nIt is:"<<totalcounter<<" out of:"<<ResVecLen<<"\n\n";

/****************/



        
    }
};



int main() {
    testbench tb( "qs.bin", "ref.bin" );
    tb.run_seq();
}
