#include <stdio.h>
#include <stdint.h>
#include <iostream>
#include <vector>
#include <oclUtils.h>

#include "CL/opencl.h"


#include "utilsPaPaRa.h"

void print_sequence(unsigned int * sequence, size_t length)
{
	for(size_t i = 0; i<length; i++)
	{
		std::cout<<"\n"<<i<<":"<<sequence[i];
	}
}

void print_sequence_TEST(unsigned int * sequence, size_t length, std::vector<int> &a, std::vector<unsigned int> &a_aux)
{
	for(size_t i = 0; i<length-32; i++)
	{
		std::cout<<"\n"<<i<<":"<<sequence[i]<<"--"<<a[i]<<"--"<<a_aux[i];
	}
}

void release_mem_object( cl_mem m ) {

	if( m != 0 ) {
		cl_int ret = clReleaseMemObject( m );
		ocl_check(ret, CL_SUCCESS);
	}
}


cl_mem allocatePinnedHostMemory_u32 (size_t memorySize, cl_context theContext, cl_command_queue theCommandQueue)
{
	cl_mem PinnedHostMemory = NULL;
	cl_int ciErrNum;
	
	PinnedHostMemory = clCreateBuffer(theContext, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, memorySize * sizeof(unsigned int), NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	return PinnedHostMemory;	
}

cl_mem allocatePinnedHostMemory_32 (size_t memorySize, cl_context theContext, cl_command_queue theCommandQueue)
{
	cl_mem PinnedHostMemory = NULL;
	cl_int ciErrNum;
	
	PinnedHostMemory = clCreateBuffer(theContext, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, memorySize * sizeof(int), NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	return PinnedHostMemory;	
}


cl_mem initializePinnedHostMemory_u32 (size_t memorySize, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue)
{
	unsigned int * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned int *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(unsigned int), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	for(size_t i = 0; i < memorySize; i++)
	{
		MappedPointer[i] = 1;
	}

	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, (void*)MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);
	MappedPointer = NULL;  

	return PinnedHostMemory;
}

cl_mem initializePinnedHostMemory_u32_Sequence (size_t memorySize, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue, unsigned int * input)
{
	unsigned int * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned int *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(unsigned int), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	for(size_t i = 0; i < memorySize; i++)
	{
		MappedPointer[i] = input[i];
	}

	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, (void*)MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);
	MappedPointer = NULL;  

	return PinnedHostMemory;
}

cl_mem initializePinnedHostMemory_32_Sequence (size_t memorySize, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue, int * input)
{
	int * MappedPointer = NULL;
	cl_int ciErrNum;

        MappedPointer = (int *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(int), 0, NULL, NULL, &ciErrNum);
        ocl_check(ciErrNum, CL_SUCCESS);
 
        for(size_t i = 0; i < memorySize; i++)
        {
            MappedPointer[i] = input[i];
        }
	
        ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, (void*)MappedPointer, 0, NULL, NULL);
        ocl_check(ciErrNum, CL_SUCCESS);
	MappedPointer = NULL;  

	return PinnedHostMemory;
}

cl_mem initializePinnedHostMemory_u32_specialArguments (size_t memorySize, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue, unsigned int size)
{
	unsigned int * MappedPointer = NULL;
	cl_int ciErrNum;

        MappedPointer = (unsigned int *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(unsigned int), 0, NULL, NULL, &ciErrNum);
        ocl_check(ciErrNum, CL_SUCCESS);
 
      //  for(size_t i = 0; i < memorySize; i++)
       // {
       //     MappedPointer[i] = 1;
      //  }
	MappedPointer[0] = 8800;
	MappedPointer[1] = 98;
	MappedPointer[2] = size;

	
        ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, (void*)MappedPointer, 0, NULL, NULL);
        ocl_check(ciErrNum, CL_SUCCESS);
	MappedPointer = NULL;  

	return PinnedHostMemory;
}

cl_mem allocateDeviceMemory_u32 (size_t memorySize, cl_context theContext)
{
	cl_mem DeviceMemory = NULL;
	cl_int ciErrNum;

	DeviceMemory = clCreateBuffer(theContext, CL_MEM_READ_WRITE, memorySize * sizeof(unsigned int), NULL, &ciErrNum);
    	ocl_check(ciErrNum, CL_SUCCESS);

	return DeviceMemory;
}

cl_mem allocateDeviceMemory_32 (size_t memorySize, cl_context theContext)
{
	cl_mem DeviceMemory = NULL;
	cl_int ciErrNum;

	DeviceMemory = clCreateBuffer(theContext, CL_MEM_READ_WRITE, memorySize * sizeof(int), NULL, &ciErrNum);
    	ocl_check(ciErrNum, CL_SUCCESS);

	return DeviceMemory;
}

void copyPinnedHostMemoryToDeviceMemory_u32 (size_t memorySize, cl_mem PinnedHostMemory, cl_mem DeviceMemory, cl_command_queue theCommandQueue)
{
	unsigned int * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned int *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_READ, 0, memorySize * sizeof(unsigned int), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);
	
	ciErrNum = clEnqueueWriteBuffer(theCommandQueue, DeviceMemory, CL_FALSE, 0, memorySize * sizeof(unsigned int), MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);

	ciErrNum = clFinish(theCommandQueue);
	ocl_check(ciErrNum, CL_SUCCESS);

	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, MappedPointer, 0, 0, 0 );
	ocl_check(ciErrNum, CL_SUCCESS);
}

void copyPinnedHostMemoryToDeviceMemory_32 (size_t memorySize, cl_mem PinnedHostMemory, cl_mem DeviceMemory, cl_command_queue theCommandQueue)
{
	int * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (int *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_READ, 0, memorySize * sizeof(int), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);
	
	ciErrNum = clEnqueueWriteBuffer(theCommandQueue, DeviceMemory, CL_FALSE, 0, memorySize * sizeof(int), MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);

	ciErrNum = clFinish(theCommandQueue);
	ocl_check(ciErrNum, CL_SUCCESS);


	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, MappedPointer, 0, 0, 0 );
	ocl_check(ciErrNum, CL_SUCCESS);
}

unsigned int * copyDeviceMemoryToPinnedHostMemory_u32 (size_t memorySize, cl_mem DeviceMemory, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue, bool auto_unpin)
{
	unsigned int * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned int *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(unsigned int), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	ciErrNum = clEnqueueReadBuffer(theCommandQueue, DeviceMemory, CL_FALSE, 0, memorySize * sizeof(unsigned int), MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);

	if( auto_unpin ) {
		clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, MappedPointer, 0, 0, 0 );
		MappedPointer = 0;
	}


	return MappedPointer;	
}

int * copyDeviceMemoryToPinnedHostMemory_32 (size_t memorySize, cl_mem DeviceMemory, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue, bool auto_unpin)
{
	int * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (int *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(int), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	ciErrNum = clEnqueueReadBuffer(theCommandQueue, DeviceMemory, CL_FALSE, 0, memorySize * sizeof(int), MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);

	if( auto_unpin ) {
		clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, MappedPointer, 0, 0, 0 );
		MappedPointer = 0;
	}


	return MappedPointer;	
}


/*****************************************************************************************************************************/

cl_mem allocatePinnedHostMemory (size_t memorySize, cl_context theContext, cl_command_queue theCommandQueue)
{
	cl_mem PinnedHostMemory = NULL;
	cl_int ciErrNum;
	
	PinnedHostMemory = clCreateBuffer(theContext, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, memorySize * sizeof(unsigned char), NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	return PinnedHostMemory;	
}

cl_mem initializePinnedHostMemory (size_t memorySize, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue)
{
	unsigned char * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned char *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(unsigned char), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	for(size_t i = 0; i < memorySize; i++)
	{
		MappedPointer[i] = 2;
	}

	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, (void*)MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);
	MappedPointer = NULL;  

	return PinnedHostMemory;
}

cl_mem initializePinnedHostMemoryPaPaRaSpecial (size_t memorySize, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue, std::vector<int> &ref_vec, std::vector<unsigned int> &ref_aux)
{
	unsigned char * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned char *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(unsigned char), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	for(size_t i = 0; i < memorySize; i++)
	{
		MappedPointer[i] = (unsigned char)(ref_vec[i]<<1)+ref_aux[i];
	}

	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, (void*)MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);
	MappedPointer = NULL;  

	return PinnedHostMemory;
}

cl_mem copyToPositionPinnedHostMemoryPaPaRaSpecial (size_t position, size_t entry_size_cur, size_t entry_size_max, size_t memorySize, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue, std::vector<unsigned char> &input_vec)
{
	unsigned char * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned char *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(unsigned char), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	for(size_t i = position; i < position + entry_size_max; i++)
	{
		if (i < position + entry_size_cur)
		{
			MappedPointer[i] = input_vec[i];
		}
		else
		{
			MappedPointer[i] = 0;
		}
	}

	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, (void*)MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);
	MappedPointer = NULL;  

	return PinnedHostMemory;
}

cl_mem allocateDeviceMemory (size_t memorySize, cl_context theContext)
{
	cl_mem DeviceMemory = NULL;
	cl_int ciErrNum;

	DeviceMemory = clCreateBuffer(theContext, CL_MEM_READ_WRITE, memorySize * sizeof(unsigned char), NULL, &ciErrNum);
    	ocl_check(ciErrNum, CL_SUCCESS);

	return DeviceMemory;
}

void copyPinnedHostMemoryToDeviceMemory (size_t memorySize, cl_mem PinnedHostMemory, cl_mem DeviceMemory, cl_command_queue theCommandQueue)
{
	unsigned char * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned char *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_READ, 0, memorySize * sizeof(unsigned char), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);
	
	ciErrNum = clEnqueueWriteBuffer(theCommandQueue, DeviceMemory, CL_FALSE, 0, memorySize * sizeof(unsigned char), MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);

	ciErrNum = clFinish(theCommandQueue);
	ocl_check(ciErrNum, CL_SUCCESS);
}

unsigned char * copyDeviceMemoryToPinnedHostMemory (size_t memorySize, cl_mem DeviceMemory, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue)
{
	unsigned char * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned char *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_WRITE, 0, memorySize * sizeof(unsigned char), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	ciErrNum = clEnqueueReadBuffer(theCommandQueue, DeviceMemory, CL_FALSE, 0, memorySize * sizeof(unsigned char), MappedPointer, 0, NULL, NULL);
	ocl_check(ciErrNum, CL_SUCCESS);

	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, MappedPointer, 0, 0, 0 );
	ocl_check(ciErrNum, CL_SUCCESS);
	return MappedPointer;	
}

void printPinnedHostMemoryContent (size_t memorySize, cl_mem PinnedHostMemory, cl_command_queue theCommandQueue)
{
	unsigned char * MappedPointer = NULL;
	cl_int ciErrNum;

	MappedPointer = (unsigned char *) clEnqueueMapBuffer(theCommandQueue, PinnedHostMemory, CL_TRUE, CL_MAP_READ, 0, memorySize * sizeof(unsigned char), 0, NULL, NULL, &ciErrNum);
	ocl_check(ciErrNum, CL_SUCCESS);

	//for(size_t i = 0; i < memorySize; i++)
	//{
	//printf("%d:%u\n",10,MappedPointer[10]);
	//}

	ciErrNum = clEnqueueUnmapMemObject(theCommandQueue, PinnedHostMemory, MappedPointer, 0, 0, 0 );
	ocl_check(ciErrNum, CL_SUCCESS);
}







void test()
{
  printf("Test");
}


char * malloc_CHAR_PINNED(cl_context context, cl_command_queue command_queue, cl_mem memobj, unsigned int size)
{
	char * memhost = NULL;
	cl_int err;

	// Create host buffer
	memobj = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size, NULL, &err);	
	// Error check
	//TODO
	
	// Get mapped pointer
	memhost = (char*)clEnqueueMapBuffer(command_queue, memobj, CL_TRUE, CL_MAP_WRITE, 0, size, 0, NULL, NULL, &err);
	// Error check
	//TODO

	// Initialize memory
	for(unsigned int i = 0; i < size; i++)
        {
            memhost[i] = 0;
        }

	// Unmap and validate
	err = clEnqueueUnmapMemObject(command_queue, memobj, memhost, 0, NULL, NULL);
	// Error check
	//TODO	

	return memhost;
}

cl_mem uint8_memallocPINNED (cl_context context, cl_command_queue command_queue, unsigned int size)
{
	cl_mem memobj = NULL;
	cl_int err;

	// Create host buffer
	memobj = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size, NULL, &err);	
	// Error check
	//TODO
	
	// Get mapped pointer
	//memhost = (uint8_t*)clEnqueueMapBuffer(command_queue, *memobj, CL_TRUE, CL_MAP_WRITE, 0, size, 0, NULL, NULL, &err);
	// Error check
	//TODO

	// Initialize memory
	//for(size_t i = 0; i < size; i++)
        //{
            //memhost[i] = 0;
        //}

	// Unmap and validate
        //err = clEnqueueUnmapMemObject(command_queue, *memobj, (void*)memhost, 0, NULL, NULL);
	// Error check
	//TODO	

	return memobj;
}

uint8_t * uint8_meminitPINNED_special(cl_context context, cl_command_queue command_queue, cl_mem memobj, unsigned int size, std::vector<int> &ref_vec, std::vector<unsigned int> &ref_aux)
{
	cl_int err;

	uint8_t * memhost = NULL;
	
	// Get mapped pointer
	memhost = (uint8_t *)clEnqueueMapBuffer(command_queue, memobj, CL_TRUE, CL_MAP_WRITE, 0, size, 0, NULL, NULL, &err);
	clFlush(command_queue);
	// Error check
	//TODO
	if (err!=CL_SUCCESS)
	{
	printf("error: %d\n",err);
}
	// Initialize memory
	for(size_t i = 0; i < size; i++)
        {
            memhost[i] = (uint8_t)(ref_vec[i]<<1)+ref_aux[i];
            //printf("%d: %d\n",i,memhost[i]);
        }

	// Unmap and validate
        err = clEnqueueUnmapMemObject(command_queue, memobj, (void*)memhost, 0, NULL, NULL);
	// Error check
	//TODO	
	clFlush(command_queue);

	//memhost = (uint8_t *) malloc(sizeof(uint8_t)*size);

//	clFinish(command_queue);

	return memhost;
}

cl_mem uint8_memallocPINNED_initSPECIAL (cl_context context, cl_command_queue command_queue, unsigned int size, std::vector<int> &ref_vec, std::vector<unsigned int> &ref_aux)
{
	cl_mem memobj = NULL;
	uint8_t *memhost;
	cl_int err;

	// Create host buffer
	memobj = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size, NULL, &err);	
	// Error check
	//TODO
	
	// Get mapped pointer
	memhost = (uint8_t*)clEnqueueMapBuffer(command_queue, memobj, CL_TRUE, CL_MAP_WRITE, 0, size, 0, NULL, NULL, &err);
	// Error check
	//TODO

	for(size_t i = 0; i < size; i++)
        {
            memhost[i] = (uint8_t)(ref_vec[i]<<1)+ref_aux[i];
            //printf("%d: %d\n",i,memhost[i]);
        }

	// Unmap and validate
        err = clEnqueueUnmapMemObject(command_queue, memobj, (void*)memhost, 0, NULL, NULL);
	// Error check
	//TODO	

	return memobj;
}




