
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <iostream>
#include <stdexcept>

#include "utilsPaPaRa.h"

void ocl_check_ex(cl_int iSample, cl_int iReference, const char* cFile, const int iLine)
{
    // An error condition is defined by the sample/test value not equal to the reference
    if (iReference != iSample)
    {
        // If the sample/test value isn't equal to the ref, it's an error by defnition, so override 0 sample/test value
        iSample = (iSample == 0) ? -9999 : iSample;

        // Log the error info
        //shrLog("\n !!! Error # %i (%s) at line %i , in file %s !!!\n\n", iSample, oclErrorString(iSample), iLine, cFile);

        // Cleanup and exit, or just exit if no cleanup function pointer provided.  Use iSample (error code in this case) as process exit code.


		//shrLogEx(LOGBOTH | CLOSELOG, 0, "Exiting...\n");

        std::cerr << "opencl error: " << iSample << " at line " << iLine << " in file " << cFile << "\n";

        throw std::runtime_error( "bailing out due to ocl error");

    }
}

//* Of *//

unsigned int get_sGloIndex (unsigned int sharedMemSize, unsigned int PadBloLen)
{
	size_t sharedInts = (sharedMemSize * 1024) / 4;

	size_t maxBlocks = sharedInts / PadBloLen;
	
	return maxBlocks * PadBloLen;
}

unsigned int get_next32mult (unsigned int input)
{
	unsigned int result=0;

	unsigned int arg = 32;

	unsigned int div32 = input / arg;
	unsigned int mod32 = input % arg;
	
	result = mod32!=0?(div32+1)*arg:div32*arg;	
	
	return result;
}
unsigned int get_constSequenceLength (unsigned int sequenceLength, unsigned int memoryBlockSize, unsigned int compressEffort)
{
	unsigned int u32min = sequenceLength % compressEffort==0?sequenceLength / compressEffort : sequenceLength / compressEffort + 1;
	unsigned int u32num = u32min % memoryBlockSize==0?u32min:u32min + memoryBlockSize - u32min % memoryBlockSize;
	return u32num;
}

void putSequenceRef(unsigned int seqIndex, std::vector<int> &a, std::vector<unsigned int> &a_aux, unsigned int constSeqLen, unsigned int * sequencesAll)
{

	size_t seqi = seqIndex * constSeqLen;
	for(size_t i = 0;i<constSeqLen;i++)
	{
		if(i<a.size())
		{
			sequencesAll[i+seqi] = 0;
			sequencesAll[i+seqi] = (sequencesAll[i+seqi]<<5)+((a[i]<<1) + a_aux[i]);
		}
		else
		{
			sequencesAll[i+seqi] = 0;	
		}
	}
}
void putSequenceRef_Inter(unsigned int seqIndex, std::vector<int> &a, std::vector<unsigned int> &a_aux, unsigned int constSeqLen, unsigned int * sequencesAll, unsigned int interSeqBloLen)
{
	for(size_t i = 0;i<constSeqLen;i++)
	{
		size_t lptr = i * interSeqBloLen + seqIndex;
		sequencesAll[lptr] = 0;
		sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + a_aux[i]);
	}
}

void putSequenceRef_InterInt(unsigned int seqIndex, std::vector<int> &a, std::vector<unsigned int> &a_aux, unsigned int constSeqLen, int * sequencesAll, unsigned int interSeqBloLen)
{
	for(size_t i = 0;i<constSeqLen;i++)
	{
		size_t lptr = i * interSeqBloLen + seqIndex;
		sequencesAll[lptr] = 0;
		//sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + a_aux[i]);
		if(a_aux[i]==1)
		{
			sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + 0);
		}
		else
		{
			sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + 1);
		}
	}
}




void putSequenceRef_CombinedS1(unsigned int seqIndex, const std::vector<int> &a, const std::vector<unsigned int> &a_aux, unsigned int constSeqLen, int * sequencesAll, unsigned int interSeqBloLen)
{
	for(size_t i = 0;i<constSeqLen;i++)
	{
		size_t lptr = i * interSeqBloLen + seqIndex;
		sequencesAll[lptr] = 0;
		//sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + a_aux[i]);
		if(a_aux[i]==1)
		{
			sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + 0);
		}
		else
		{
			sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + 1);
		}
	}
}

void putSequenceRef_CombinedS2(unsigned int seqIndex, const std::vector<int> &a, const std::vector<unsigned int> &a_aux, unsigned int constSeqLen, int * sequencesAll, unsigned int interSeqBloLen, int quelenmax, int blockiterations, int blocksizecompressed, int startpos)
{
	for(size_t i = 0;i<constSeqLen;i++)
	{
		size_t lptr = (i+quelenmax+blockiterations*blocksizecompressed) * interSeqBloLen + seqIndex;
		sequencesAll[lptr] = 0;
		//sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + a_aux[i]);
		if(a_aux[i+startpos]==1)
		{
			sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i+startpos]<<1) + 0);
		}
		else
		{
			sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i+startpos]<<1) + 1);
		}
	}
}




void putSequenceRef_InterIntComp (unsigned int seqIndex, const std::vector<int> &a, const std::vector<unsigned int> &a_aux, unsigned int constSeqLen, int start_pos, int last_pos, int * CompsequencesAll, unsigned int interSeqBloLen)
{
	const int effort = 6;
	const int llength = (last_pos - start_pos) / effort;
	size_t refindex=0;
	size_t compindex=0;

	for(size_t i=0;i<llength;i++)
	{
		compindex = i * interSeqBloLen + seqIndex;
		CompsequencesAll[compindex]=0;
		
		

		for(size_t j=0;j<effort;j++)
		{
			refindex = start_pos + i * effort + j;

			if(a_aux[refindex]==1)
			{
				CompsequencesAll[compindex] = (CompsequencesAll[compindex]<<5)+((a[refindex]<<1) + 0);
			}
			else
			{
				CompsequencesAll[compindex] = (CompsequencesAll[compindex]<<5)+((a[refindex]<<1) + 1);
			}
		}


				
	}

}


void putSequenceRef_InterIntCompCombined (unsigned int seqIndex, const std::vector<int> &a, const std::vector<unsigned int> &a_aux, unsigned int constSeqLen, int start_pos, int last_pos, int * CompsequencesAll, unsigned int interSeqBloLen, unsigned int startpos)
{
	const int effort = 6;
	const int llength = (last_pos - start_pos) / effort;
	size_t refindex=0;
	size_t compindex=0;

	for(size_t i=0;i<llength;i++)
	{
		compindex = (i+startpos) * interSeqBloLen + seqIndex;
		CompsequencesAll[compindex]=0;
		
		

		for(size_t j=0;j<effort;j++)
		{
			refindex = start_pos + i * effort + j;

			if(a_aux[refindex]==1)
			{
				CompsequencesAll[compindex] = (CompsequencesAll[compindex]<<5)+((a[refindex]<<1) + 0);
			}
			else
			{
				CompsequencesAll[compindex] = (CompsequencesAll[compindex]<<5)+((a[refindex]<<1) + 1);
			}
		}				
	}
}

void putSequenceRef_InterIntCompRev (unsigned int seqIndex, std::vector<int> &a, std::vector<unsigned int> &a_aux, unsigned int constSeqLen, int start_pos, int last_pos, int * CompsequencesAll, unsigned int interSeqBloLen)
{
	const int effort = 6;
	const int llength = (last_pos - start_pos) / effort;
	size_t refindex=0;
	size_t compindex=0;
	

	for(size_t i=0;i<llength;i++)
	{
		compindex = i * interSeqBloLen + seqIndex;
		CompsequencesAll[compindex]=0;
		
		

		for(int j=effort-1;j>=0;j--)
		{
			refindex = start_pos + i * effort + j;

			if(a_aux[refindex]==1)
			{
				CompsequencesAll[compindex] = (CompsequencesAll[compindex]<<5)+((a[refindex]<<1) + 0);
			}
			else
			{
				CompsequencesAll[compindex] = (CompsequencesAll[compindex]<<5)+((a[refindex]<<1) + 1);
			}
		}

				
	}

}


void putSequenceRef_Inter_Dual(unsigned int seqIndex, std::vector<int> &a, std::vector<unsigned int> &a_aux, unsigned int constSeqLen, int * sequencesAll_a, unsigned int * sequencesAll_a_aux, unsigned int interSeqBloLen)
{
	for(size_t i = 0;i<constSeqLen;i++)
	{
		size_t lptr = i * interSeqBloLen + seqIndex;
		//sequencesAll[lptr] = 0;
		//sequencesAll[lptr] = (sequencesAll[lptr]<<5)+((a[i]<<1) + a_aux[i]);

		sequencesAll_a[lptr] = a[i];
		if(a_aux[i]==1)
		{
			sequencesAll_a_aux[lptr] = 0;
		}
		else
		{
			sequencesAll_a_aux[lptr] = 1;
		}
		//sequencesAll_a_aux[lptr] = a_aux[i];
	}
}

void putSequenceQue(unsigned int seqIndex, const std::vector<unsigned char> &a,  unsigned int constSeqLen, unsigned int * sequencesAll)
{

	size_t seqi = seqIndex * constSeqLen;
	for(size_t i = 0;i<constSeqLen;i++)
	{
		if(i<a.size())
		{
			sequencesAll[i+seqi] = a[i];
		}
		else
		{
			sequencesAll[i+seqi] = 0;	
		}
	}
}
//float get_gMemTotal(unsigned int num_of_queries, unsigned int num_of_references, unsigned int constQueryLength, unsigned int constReferenceLength)
//{
//	size_t integers = num_of_queries * constQueryLength + num_of_references * constReferenceLength;
//
//	float bytes = integers / 4;
//
//	float GBs = bytes / 1024 / 1024 / 1024;
//
//	return GBs;
//}

float get_shareMemLineSize (unsigned int constSequenceLength)
{
	int bytes = constSequenceLength * 4;
	return (float)bytes / 1024.0;
}
//***//


void showBitsMSB(unsigned int input)
{
	unsigned int vec[32];
	unsigned int tinput=input;
	printf("\n");
	for(size_t i = 32; i>0 ; i--)
	{
		vec[i-1]=(unsigned int)tinput%2;
		tinput = tinput / 2;
	}
	for(size_t i = 0; i<32; i++)
	{
		printf("%u",vec[i]);
	}
	
}

/*************************************************************************************************************/
const size_t getCompactRefSize (const size_t compactRefMode, const size_t reflen, const size_t alignBlockSize)
{
	size_t u32min = reflen % compactRefMode==0?reflen / compactRefMode : reflen / compactRefMode + 1;
	size_t u32num = u32min % alignBlockSize==0?u32min:u32min + alignBlockSize - u32min % alignBlockSize;
	return u32num;
}
/*************************************************************************************************************/
const size_t getCompactSize (const size_t compactRefMode, const size_t reflen, const size_t alignBlockSize)
{
	size_t u32min = reflen % compactRefMode==0?reflen / compactRefMode : reflen / compactRefMode + 1;
	size_t u32num = u32min % alignBlockSize==0?u32min:u32min + alignBlockSize - u32min % alignBlockSize;
	return u32num;
}
/*************************************************************************************************************/
unsigned int * getCompactReference (const size_t compactRefMode, const size_t reflen, std::vector<int> &a, std::vector<unsigned int> &a_aux,const size_t compactRefSize)
{
	unsigned int * compactRef = (unsigned int *) malloc(sizeof(unsigned int) * compactRefSize);
	
	size_t t_c=0;
	size_t ref_i=0;
	unsigned int tmp = 0;

	for(size_t i=0;i<compactRefSize;i++)
	{		
		for(size_t t_c=0;t_c<compactRefMode;t_c++)
		{
			
			if(ref_i!=reflen)
			{
			   tmp = (tmp<<5) + ((a[ref_i]<<1)+a_aux[ref_i]);
			   ref_i++;
			}
			else
			{
			   tmp = (tmp<<5);
			}
			if(t_c==compactRefMode-1)
			{
				compactRef[i]=tmp;
				tmp=0;
			}
		}		
	}

	return compactRef;
}
/*************************************************************************************************************/

unsigned int * getCompactQuery (const size_t compactQueryMode, const size_t quelen, std::vector<unsigned char> &a,const size_t compactQuerySize)
{
	unsigned int * compactQue = (unsigned int *) malloc(sizeof(unsigned int) * compactQuerySize);
	
	size_t t_c=0;
	size_t que_i=0;
	unsigned int tmp = 0;

	for(size_t i=0;i<compactQuerySize;i++)
	{		
		for(size_t t_c=0;t_c<compactQueryMode;t_c++)
		{
			
			if(que_i!=quelen)
			{
			   tmp = (tmp<<4) + a[que_i];
			   que_i++;
			}
			else
			{
			   tmp = (tmp<<4);
			}
			if(t_c==compactQueryMode-1)
			{
				compactQue[i]=tmp;
				tmp=0;
			}
		}		
	}

	return compactQue;
}
/*************************************************************************************************************/




unsigned int fixthis (size_t ref_len, size_t sharedKBytes, size_t memBanks)
{
	
	unsigned int u32_perRef = ref_len / 6; // TODO: This needs to be changed for per-thread pseudo-vectorization experiments.
	
	unsigned int u32_shared = sharedKBytes * 1024 / 4;

	unsigned int u32Ref_perComputeBlock = u32_perRef * memBanks;

	// TODO: Find max query length. calculate max reference length for 1 fill. Go for inter-query parallelization
	
	printf("MAX ref len: %u one ref size = %u \n ComputeBlock size = %u \n total avail: %u max_refs: %u\n",u32_shared * 6, u32_perRef, u32Ref_perComputeBlock, u32_shared, u32_shared / u32_perRef);

	//unsigned int  
	
	return 0;
	
}

unsigned int * refInpAll_Reorg (size_t ref_len, size_t refAll_num, std::vector<std::vector <int> > inpRefAll, int alignBlock_len)
{
	size_t block_sz = refAll_num % alignBlock_len==0?refAll_num:refAll_num + alignBlock_len - refAll_num % alignBlock_len ;

	unsigned int * refAll_mem = (unsigned int *) malloc (sizeof(unsigned int)* (ref_len * block_sz));
	
	size_t refAll_ptr = 0;

	for( size_t i = 0; i < ref_len; i++)
	{
	    for( size_t j = 0; j < block_sz; j++ ) 
	    {  	
		if (j<refAll_num) { refAll_mem[refAll_ptr++] = (unsigned int) ((inpRefAll[j][i])); }
		else		  { refAll_mem[refAll_ptr++] = 0 ; }			
            }
	}
	
	return refAll_mem;	
}


